/* --------------------------------------------------------------- */
/* tollosAcc.c -- accelerometer functions for Tollos               */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010.  All rights reserved.       */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* These functions provide basic accelerometer functions for       */
/* Tollos, intended to be device-independent (although at the      */
/* moment an i2c connection is assumed).                           */
/*                                                                 */
/* To use, the device must be initialized using accInit.  This     */
/* indicates which i2c channel the device is on and which port/pin */
/* is used for DataReady interrupt.  By default the accelerometer  */
/* range will be at least 2g; this may be changed with accRange.   */
/*                                                                 */
/* The latest accelerometer reading can be read safely using       */
/* accGetXYZ, which returns the three axes in milli-g.  The axes   */
/* directions are as defined by the device manufacturer.           */
/*                                                                 */
/* The update rate is selected by the device driver code and will  */
/* be at least 25Hz (that is, the data will be no more than 40ms   */
/* old).  Update is not necessarily synchronous with the Tollos    */
/* tick rate.                                                      */
/* --------------------------------------------------------------- */

#define  TOLLOS_IO            // allow use of I/O functions
#include "tollos.h"

// vector of readings from Accelerometer
typedef struct {
  int x;                      // X reading [mg]
  int y;                      // Y reading [mg]
  int z;                      // Z reading [mg]
  } accData;

// double-buffer of readings
static accData readbuff[2];   // double buffer
static uint    goodbuff=0;    // which buffer has safe data to read

// state
#define ACC_STOP  0           // stopped (not yet initialized)
#define ACC_WAIT  1           // waiting for good data
#define ACC_GOOD  2           // have readable data
#define ACC_GOOD2 3           // have two readable data
static uint state=ACC_STOP;   // current state

// local function
static void accHandler(void); // interrupt handler

/* --------------------------------------------------------------- */
/* accGetXYZ -- read latest x, y, and z accelerations              */
/*                                                                 */
/*   x, y, z -- receive accelerations [mg: 1000 = 1.000 g], or all */
/*     zero if error                                               */
/*                                                                 */
/*   returns 0 if data good, -1 if error (no good data yet, or     */
/*     device not initialized)                                     */
/*                                                                 */
/* Results are returned scaled to milli-gs (1000 = 1g), without    */
/* any calibration adjustments.                                    */
/* --------------------------------------------------------------- */
int accGetXYZ(int *x, int *y, int*z) {

  if (state==ACC_GOOD) state=ACC_WAIT;       // bump down state
   else if (state==ACC_GOOD2) state=ACC_GOOD;
   else {                                    // ACC_STOP or ACC_WAIT
    *x=0; *y=0; *z=0;
    return -1;
    }

  // return data from the 'good' buffer
  *x=readbuff[goodbuff].x;
  *y=readbuff[goodbuff].y;
  *z=readbuff[goodbuff].z;

  return 0;                   // is good
  } // accGetXYZ

/* --------------------------------------------------------------- */
/* accHandler -- interrupt handler, called when Data Ready         */
/*                                                                 */
/* This reads the data into the 'bad' (dirty) buffer, then when    */
/* good sets the goodbuff to point to it and indicates good data.  */
/* --------------------------------------------------------------- */
static void accHandler(void) {
  int rc;                               // work
  uint dirty=goodbuff ? 0 : 1;          // buffer to read into

  rc=accGetXYZIO(&readbuff[dirty].x, &readbuff[dirty].y, &readbuff[dirty].z);
  // interrupt is cleared by the read
  if (rc<0) return;                     // bad data

  goodbuff=dirty;                       // switch
  if (state==ACC_WAIT) state=ACC_GOOD;  // bump up state
   else if (state==ACC_GOOD) state=ACC_GOOD2;
  } // accHandler

/* --------------------------------------------------------------- */
/* accInit -- configure and initialize accelerometer               */
/*                                                                 */
/*   channel -- i2c channel on which device is to be found         */
/*   drport  -- port for the Data Ready interrupt pin              */
/*   drpin   -- pin number (not bit mask) for the Data Ready       */
/*     interrupt pin                                               */
/*                                                                 */
/*   returns 0 if initialized OK, negative if error                */
/*                                                                 */
/* This sets up the pin and interrupt handler for the Data Ready   */
/* interrupt from the device then calls the device driver          */
/* initialization function to start the device.                    */
/*                                                                 */
/* The accInitIO function will set:                                */
/* * an update rate of at least 25 Hz (but not necessarily higher  */
/* than that, to save power)                                       */
/* * a range of at least 2g (call accRange if a larger range is    */
/* needed)                                                         */
/* * an interrupt on Data Ready.                                   */
/*                                                                 */
/* Note that it can be 40ms or more before the first valid data    */
/* are available, so an accGetXYX call before that time may report */
/* no data available.                                              */
/* --------------------------------------------------------------- */
int accInit(uint channel, uint drport, uint drpin) {
  int rc;                     // work
  state=ACC_STOP;

  // set up interrupt pin from number (bitFuns use a mask)
  rc=bitEnable(drport, 1<<drpin, BIT_RISE, accHandler);

  if (rc==0) rc=accInitIO(channel);
  if (rc==0) state=ACC_WAIT;  // data should arrive soon
  return rc;
  } // accInit

/* --------------------------------------------------------------- */
/* accRange -- change range of accelerometer                       */
/*                                                                 */
/*   range -- range requested in milli-g [1000 = 1 g]              */
/*                                                                 */
/*   returns 0 if range changed OK, -1 if error (range too high)   */
/*                                                                 */
/* The accelerometer will be set to a smallest available range     */
/* that is >= the requested range, to maximize the accuracy of     */
/* returned results.                                               */
/* --------------------------------------------------------------- */
int accRange(uint range) {
  return accRangeIO(range);   // pass on and back
  } // accRange

/* --------------------------------------------------------------- */
/* accStop -- stop and power-down accelerometer                    */
/*                                                                 */
/*   returns 0 if accelerometer stopped OK, negative if an error   */
/*     or it was not running.                                      */
/* --------------------------------------------------------------- */
int accStop(void) {
  state=ACC_STOP;             // be clean
  return accStopIO();         // pass on and back
  } // accStop

